# frozen_string_literal: true

require 'rails_helper'

# %w[
#   applicant
#   housing_request
# ].each { |factory_name|require Hermes::Engine.root.join('spec', 'factories', "#{factory_name.pluralize}_factory").to_s }

RSpec.describe Hermes::HousingRequest, type: :model do
  describe 'basic controls' do
    let!(:user1) {create(:user)}
    let!(:applicant1) do
      create(
        :applicant,
        user: user1,
        monthly_charges: {}
      )
    end
    
    it 'has a valid factory' do
      expect(FactoryBot.build(:housing_request, main_applicant: applicant1)).to be_valid
    end
  end
  
  context 'with simple setup' do
    let!(:user1) {create(:user)}
    let!(:user2) {create(:user)}
    let!(:user3) {create(:user)}
    let!(:applicant1) {
      create(
        :applicant,
        user: user1,
        monthly_charges: {}
      )
    }
    let!(:applicant2) {
      create(
        :applicant,
        user: user2
      )
    }
    let!(:applicant3) {
      create(
        :applicant,
        user: user3
      )
    }
    
    let!(:housing_request1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1
      )
    end
    
    let!(:housing_request2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1,
        coapplicants: [
          {rel_type: 'Conjoint (e)', id: applicant2.id.to_s},
          {rel_type: 'Parent (e)', id: applicant3.id.to_s}
        ]
      )
    end
    
    describe 'Eligibility query are valid' do
      it 'should include right typology ranges' do
        typology_range1 = housing_request1.calculate_typology_range
        typology_range2 = housing_request2.calculate_typology_range
        range1_to_i = typology_range1.map {|typology| Integer(typology.sub('T', ''))}
        range2_to_i = typology_range2.map {|typology| Integer(typology.sub('T', ''))}
        
        (1..2).each do |n|
          (1..2).each do |i|
            expect(binding.local_variable_get("range#{n}_to_i".to_sym)).to include i
          end
        end
        expect(range1_to_i.max).to be 2
        expect(range2_to_i).to include 3
        expect(range2_to_i.max).to be 3
      end
      
      it 'eligibility should incorpore typology' do
        typology_filter1 = housing_request1.typology_filter
        expect(typology_filter1.first[:$or][0][:typology]).to eq 'T1'
        expect(typology_filter1.first[:$or][1][:typology]).to eq 'T2'
      end
      
      describe 'edit action dont duplicate coapplicants' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.seeking_elevator = false
          housing_request2.save
          housing_request2.reload
          housing_request2.coapplicants = [{rel_type: 'M', id: applicant3.id.to_s}]
          housing_request2.save
          applicant3.destroy
          housing_request2.reload
          housing_request2.forcing_duplicate = true
          housing_request2.save
          housing_request2.reload
        end
        it 'coapplicants doesnt change' do
          expect(housing_request2.reload.applicant_ids.length).to eq(0)
        end
      end
      
      describe 'delete applicantId delete coapplicants' do
        before :each do
          housing_request2.applicant_ids = []
          housing_request2.save
        end
        it 'coapplicants doesnt change' do
          expect(housing_request2.reload.coapplicants.length).to eq(0)
        end
      end
    end
    
    describe "#update_seeked_location" do
      let(:list) do
        {'nom': 'Fréjus', 'code': '83061', 'codeDepartement': '81', 'codesPostaux': ['83900', '83370', '']}
      end
      let(:lists) do
        [
          {'nom': 'Paris', 'code': '75001', 'codeDepartement': '75', 'codesPostaux': ['75006', nil]},
          {'nom': 'Meaux', 'code': '77000', 'codeDepartement': '77'},
          {'nom': 'Paris', 'code': '75002', 'codeDepartement': '75', 'codesPostaux': ['75006', '93000']}
        ]
      end
      let(:broken_list) do
        {'code': '77013', 'codePostal': '77013', 'foo': 'bir'}
      end
      
      it 'sets seeked fields on save' do
        housing_request1.seeked_localisation_list = [list]
        expect {housing_request1.save}.to change {housing_request1.seeked_departments}.from([]).to(['81'])
                                            .and change {housing_request1.truly_seeked_departments}.from([]).to(['81', '83'])
                                                   .and change {housing_request1.seeked_district_postal_codes}.from([]).to(['83370', '83900'])
                                                          .and change {housing_request1.seeked_districts}.from([]).to(['Fréjus'])
      end
      
      it 'ensures the sorted uniqueness of seeked fields' do
        housing_request1.seeked_localisation_list = lists + [list]
        housing_request1.save
        expect(housing_request1.seeked_departments).to eq ['75', '77', '81']
        expect(housing_request1.truly_seeked_departments).to eq ['75', '77', '81', '83', '93']
        expect(housing_request1.seeked_district_postal_codes).to eq ['75006', '83370', '83900', '93000']
        expect(housing_request1.seeked_districts).to eq ['Fréjus', 'Meaux', 'Paris']
      end
      
      it 'returns error' do
        housing_request1.seeked_localisation_list = [broken_list]
        housing_request1.save
        expect(housing_request1.errors.full_messages.first).to match /non-conform seeked_localisation_list/
      end
    end
    
    describe 'Relationships are valid' do
      it 'should has correct housing_requests situation' do
        expect(applicant1.first_active_request.id).to eq(housing_request2.id)
        expect(applicant1.housing_requests.length).to eq(2)
        expect(applicant1.housing_requests.actives.length).to eq(1)
      end
      
      it 'should has correct values' do
        revenues_sum = (applicant1.monthly_revenues || 0) + (applicant2.monthly_revenues || 0)
        expect(housing_request2.monthly_revenues).to eq(revenues_sum)
      end
      
      it 'should has correct coapplicants' do
        coapplicants = housing_request2.coapplicants
        expect(coapplicants.length).to eq(2)
        expect(coapplicants.map {|coapplicant| coapplicant[:rel_type]}).to include('Conjoint (e)').and include('Parent (e)')
        expect(coapplicants.map {|coapplicant| coapplicant[:id]}).to include(applicant2.id).and include(applicant3.id)
      end
      
      describe 'after coapplicants update' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.save
        end
        
        it 'should has updated with valid data' do
          r_request = housing_request2.reload
          expect(r_request.applicant_ids.length).to be 1
          expect(r_request.applicant_ids.first).to eq applicant3.id
          expect(r_request.coapplicants.first[:rel_type]).to eq 'Conjoint (e)'
          expect(Hermes::ApplicantLink.where(housing_request_id: r_request.id).count).to be 1
        end
      end
    end
    
    describe 'applicant should have only public offer as eligibility when last active request becomes inactive' do
      it 'should has first active request with valid data' do
        expect(housing_request2.main_applicant.housing_requests.actives.count).to eq 1
        expect(housing_request2.is_active).to be true
        expect(housing_request2.main_applicant.first_active_request.id.to_s).to eq housing_request2.id.to_s
      end
      
      context 'is_active is set to false' do
        before :each do
          housing_request2.update(is_active: false)
        end
        
        it 'should has is_active == false' do
          expect(housing_request2.reload.is_active).to be false
        end
        
        it 'should has its main_applicant.eligibility == {is_public: true}' do
          expect(housing_request2.reload.main_applicant.eligibility_query).to eq '{"is_public":true}'
        end
      end
    end
    
    describe "#generate_als_id" do
      let(:housing_request3) do
        build(
          :housing_request,
          is_active: true,
          main_applicant: applicant1
        )
      end
      it {expect {housing_request3.save}.to change {housing_request3.als_id}.from(nil).to '0000000003'}
    end
  end
  
  context 'with large quantity' do
    nb = 20
    nb.times do |i|
      let!("applicant#{i}") {create(:applicant)}
      let!("request#{i}") {create(:housing_request, main_applicant: send("applicant#{i}"))}
    end
    
    describe 'als_id is set' do
      let(:all_requests) {Hermes::HousingRequest.only(:als_id).where(als_id: {:$ne => nil})}
      
      it "has #{nb} entities" do
        expect(all_requests.count).to be nb
      end
      
      let(:als_ids) {all_requests.pluck(:als_id).uniq.compact}
      
      it 'produces valid length' do
        expect(als_ids.map(&:length)).to all(be(10))
      end
      
      it 'has no duplicates' do
        expect(
          als_ids.map {|als_id| all_requests.where(als_id: als_id).count}
        ).to all(be(1))
      end
      
      it 'increments the right way' do
        idx = (0...nb).to_a.sample
        expect(als_ids[idx]).to eq '%010d' % (idx + 1)
      end
    end
  end
  
  context 'different applicant profil' do
    let!(:user) {create(:user)}
    let!(:user2) {create(:user)}
    
    describe 'category types as expected' do
      
      it 'Profils 7 : 1 pers avec 1 enfant à charge = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          family_situation: 'Célibataire',
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          dependents: [
            {
              rel_type: 'enfant à charge',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )
        
        expect(housing_request.reload.category_type).to be 3
      end
      
      it 'Profils 8 : 1 pers avec 1 enfant en droit de visite = catégorie de ménage 2' do
        applicant = create(
          :applicant,
          user: user,
          family_situation: 'Célibataire',
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          dependents: [
            {
              rel_type: 'droit de visite',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )
        
        expect(housing_request.reload.category_type).to be 2
      end
      
      it 'Profil 9 : 2 pers avec 1 enfant en droit de visite = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          monthly_charges: {}
        )
        coapp = create(
          :applicant,
          user: user2,
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          coapplicants: [
            {rel_type: 'Marié (e)', id: coapp.id.to_s}
          ],
          dependents: [
            {
              rel_type: 'droit de visite',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )
        
        expect(housing_request.reload.category_type).to be 3
      end
      
      it 'Profils 10 : 1 pers avec 1 enfant en garde alternée = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          family_situation: 'Célibataire',
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          dependents: [
            {
              rel_type: 'garde alternée',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )
        
        expect(housing_request.reload.category_type).to be 3
      end
      
      it 'Profil 11 : 2 pers avec 1 enfant à charge = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          monthly_charges: {}
        )
        coapp = create(
          :applicant,
          user: user2,
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          coapplicants: [
            {rel_type: 'Marié (e)', id: coapp.id.to_s}
          ],
          dependents: [
            {
              rel_type: 'enfant à charge',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )
        
        expect(housing_request.reload.category_type).to be 3
      end
    
    end
    describe "untensed_areas" do
      let!(:user1) {create(:user)}
      
      let!(:applicant1) do
        create(
          :applicant,
          user: user1,
          nationality: "1",
          monthly_charges: {}
        )
      end
      
      let!(:housing_request1) do
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant1,
          unborn_children: [{
                               'anticipated_birth' => '2020-03-23T00:00:00.000Z'
                             }],
          seeked_localisation_list: [
          ],
        )
      end
      let!(:housing_request2) do
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant1,
          seeked_localisation_list: [{"nom" => "Paris",
                                      "code" => "13201",
                                      "codesPostaux" => ['13201'],
                                      'codeDepartement' => '13',
                                      "_score" => 1}],
        )
      end
  
      let!(:untensed_area2) do
        create(
          :untensed_area,
          insee_code: "33201"
        )
      end
      describe "update on create" do
      it "set untensed area" do
        housing_request1.seeked_localisation_list = [{"nom" => "Paris",
                                                      "code" => "33201",
                                                      "codesPostaux" => ['33201'],
                                                      'codeDepartement' => '33',
                                                      "_score" => 1}]
        housing_request1.save
        
        expect(housing_request1.reload.is_untensed_area).to be true
        expect(housing_request2.reload.is_untensed_area).to be false
    
        untensed_area2.insee_code = "13201"
        untensed_area2.save
       
        perform_enqueued_jobs do
          enqueued_jobs.each do |job|
            if job[:queue] == "updateuntensedlow"
       
            job[:job].perform_now(*job[:args])
              end
          end
        end
        expect(housing_request1.reload.is_untensed_area).to be false
      end
      end
      describe "update on create" do
        let!(:untensed_area3) do
          create(
            :untensed_area,
            insee_code: "13201"
          )
        end
        it "update" do
          expect(housing_request2.reload.is_untensed_area).to be false
          perform_enqueued_jobs do
            enqueued_jobs.each do |job|
              job[:job].perform_now(*job[:args])
            end
          end
          expect(housing_request2.reload.is_untensed_area).to be true
          expect(housing_request2.housing_request_folder.get_labels.length).to be 1
        end
      end
      describe "update on create" do

        it "update" do
          untensed_area2.insee_code = "13201"
          untensed_area2.save
          expect(housing_request2.reload.is_untensed_area).to be false
          perform_enqueued_jobs do
            enqueued_jobs.each do |job|
              job[:job].perform_now(*job[:args])
            end
          end
          expect(housing_request2.reload.is_untensed_area).to be true
          expect(housing_request2.housing_request_folder.get_labels.length).to be 1
        end
      end
    end
  end
  describe "update_from_sne" do
  
    let!(:user1) {create(:user)}
  
    let!(:applicant1) do
      create(
        :applicant,
        user: user1,
        nationality: "1",
        monthly_charges: {},
        first_name: "JOSSELINE",
        last_name: "LIBAR",
        birth_date: Date.parse("1954-05-06")
      )
    end
  
    let!(:housing_request1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1,
        unborn_children: [{
                            'anticipated_birth' => '2020-03-23T00:00:00.000Z'
                          }],
        seeked_localisation_list: [
        ],
        )
    end
    let!(:nud1) do
      create(
        :nud,
        housing_request: housing_request1,
        department: '12',
        creation_date: '2016-23-14T14:53:22',
        unique_number: "1110220019793G1175"
      )
    end
    
    it "get_last_xml_from_sne" do
      require 'vcr_config'
      VCR.use_cassette("find_nud_task2", :match_requests_on => [:method, :uri, :path]) do
      
        housing_request1.get_last_xml_from_sne
        housing_request1.save
        housing_request1.update_request_from_sne
        expect(housing_request1.main_applicant.reload.amount_year_minus_two).to eq(5326.0)
      end
      
    end
  end

  describe "After actions" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor) }
    let!(:applicant) { create(:applicant, user: user, entity: entity) }  
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant: applicant, request_status: 'COMPLETE') }
    
    it "should create public shares if request is applicated" do
      hrs = Hermes::HousingRequestShare.find_by(housing_request_id: housing_request.id, lessor_id: lessor.id)
      expect(hrs).not_to be_nil
    end
  end

  describe "Entity validations" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor) }
    let!(:lessor2) { create(:lessor) }
    let!(:entity2) { create(:entity, lessor: lessor2, auto_request_validation: true) }
    let!(:applicant) { create(:applicant, user: user, entity: entity) }  
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant: applicant, request_status: 'COMPLETE') }
    
    it "should automaticly validate request" do
      applicant.update(entity: entity2)

      expect(housing_request.reload.entity_validated).to eq(true)
    end
    
    it "should not automaticly validate request" do
      entity2.update(auto_request_validation: false)
      applicant.update(entity: entity2)

      expect(housing_request.reload.entity_validated).to eq(false)
    end
  end
end
